/* This file is part of Idest.
   Copyright (C) 2009-2011 Sergey Poznyakoff
 
   Idest is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
 
   Idest is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with Idest.  If not, see <http://www.gnu.org/licenses/>. */

#include "idest.h"
#include <file.h>
#include <signal.h>

void
set_frame_value(struct id3_frame *frame, const struct ed_item *item)
{
	const char *value = item->v.value;
	union id3_field *field;
	enum id3_field_textencoding encoding =
		(latin1_option ? ID3_FIELD_TEXTENCODING_ISO_8859_1 
		               : ID3_FIELD_TEXTENCODING_UTF_8);
	id3_ucs4_t *ucs4;
	const char *condesc;
	
	switch (frame_to_item_id(frame->id)) {
	default:
		field = id3_frame_field(frame, 0);
		id3_field_settextencoding(field, encoding);

		if (latin1_option)
			ucs4 = id3_latin1_ucs4duplicate(
				     (const id3_latin1_t *) value);
		else
			ucs4 = id3_utf8_ucs4duplicate(
				     (const id3_utf8_t *)value);
		
		field = id3_frame_field(frame, 1);
		id3_field_setstrings(field, 1, &ucs4);
		free(ucs4);
		break;
		
	case item_comment:
		/* Field 0: Encoding */
		field = id3_frame_field(frame, 0);
		id3_field_settextencoding(field, encoding);
		/* Field 1: Language */
		field = id3_frame_field(frame, 1);
		id3_field_setlanguage(field, item->lang ? item->lang : "eng");
		/* Field 2: Short content descriptor */
		field = id3_frame_field(frame, 2);
		condesc = item->condesc ? item->condesc : "";
		if (latin1_option)
			ucs4 = id3_latin1_ucs4duplicate(
				     (const id3_latin1_t *)condesc);
		else
			ucs4 = id3_utf8_ucs4duplicate(
				     (const id3_utf8_t *)condesc);
		id3_field_setstring(field, ucs4);
		/* Field 3: Comment */
		field = id3_frame_field(frame, 3);
		if (latin1_option)
			ucs4 = id3_latin1_ucs4duplicate(
				     (const id3_latin1_t *) value);
		else
			ucs4 = id3_utf8_ucs4duplicate(
				     (const id3_utf8_t *)value);
		id3_field_setfullstring(field, ucs4);
		free(ucs4);
		break;
	}
}

void
safe_id3_file_update_and_close(struct id3_file *file)
{
	sigset_t set, oldset;

	sigemptyset(&set);
	sigaddset(&set, SIGINT);
	sigaddset(&set, SIGTERM);

	sigprocmask(SIG_BLOCK, &set, &oldset);
	id3_file_update(file);
	id3_file_close(file);
	sigprocmask(SIG_SETMASK, &oldset, NULL);
}

static void
update_frame(struct id3_tag *tag, const struct ed_item *item)
{
	struct id3_frame *frame	= id3_tag_findframe(tag, item->id, 0);
	if (!frame) {
		frame = id3_frame_new(item->id);
		if (id3_tag_attachframe(tag, frame))
			error(1, 0, "cannot attach new frame");
	} else if (strcmp(item->id, ID3_FRAME_COMMENT) == 0
		   && item->condesc) {
		/* Special handling for comments */
		int i = 0;
		
		do {
			union id3_field *field = id3_frame_field(frame, 2);
			char *s;
			
			if (field && (s = field_to_string(field, 0))
			    && strcmp(s, item->condesc) == 0)
				break;
		} while (frame = id3_tag_findframe(tag, item->id, ++i));
		if (!frame) {
			frame = id3_frame_new(item->id);
			if (id3_tag_attachframe(tag, frame))
				error(1, 0, "cannot attach new frame");
		}
	} else {
		struct id3_frame *fp;
		
		while (fp = id3_tag_findframe(tag, item->id, 1)) {
			id3_tag_detachframe(tag, fp);
			id3_frame_delete(fp);
		}
	}
	set_frame_value(frame, item);
}

static int
build_tag_options(struct id3_tag *tag, unsigned long location,
		  id3_length_t length, void *data)
{
	unsigned int ver = id3_tag_version(tag);

	if (ID3_TAG_VERSION_MAJOR(ver) == 1)
		*(unsigned int*)data |= IDEST_ID3V_1;
	else
		*(unsigned int*)data |= IDEST_ID3V_2;
	return 0;
}

void
set_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	int modified = 0;
	int vopt = version_option;
		
	file = id3_file_open(name, ID3_FILE_MODE_READWRITE);
	if (!file)
		error(1, errno, "cannot open file %s", name);
	tag = id3_file_tag(file);
	if (!tag) 
		abort(); /* FIXME */

	if (ed_list) {
		gl_list_iterator_t itr;
		const void *p;
		
		itr = gl_list_iterator(ed_list);
		while (gl_list_iterator_next(&itr, &p, NULL)) {
			const struct ed_item *item = p;
			update_frame(tag, item);
			modified |= 1;
		}
		gl_list_iterator_free(&itr);
	}

	/* FIXME */
	modified |= guile_transform(name, tag);
	
	if (convert_version == 0) {
		if (modified && !vopt) {
			if (id3_file_struct_ntags(file)) {
				id3_file_struct_iterate(file,
							build_tag_options,
							&vopt);
			} else 
				vopt = default_version_option;
		}
	} else if (id3_file_struct_ntags(file)) {
		id3_file_struct_iterate(file, build_tag_options, &vopt);
		if (vopt != convert_version) {
			vopt = convert_version;
			modified |= 1;
		}
	}

	if (modified && vopt) {
		int opts = 0;

		if (vopt & IDEST_ID3V_1)
			opts |= ID3_TAG_OPTION_ID3V1;
		if (!(vopt & IDEST_ID3V_2))
			opts |= ID3_TAG_OPTION_NO_ID3V2;
		id3_tag_options(tag,
				ID3_TAG_OPTION_ID3V1|ID3_TAG_OPTION_NO_ID3V2,
				opts);
	}
	if (modified)
		safe_id3_file_update_and_close(file);
	else
		id3_file_close(file);
}

void
del_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	
	file = id3_file_open(name, ID3_FILE_MODE_READWRITE);
	if (!file)
		error(1, errno, "cannot open file %s", name);
	tag = id3_file_tag(file);
	if (!tag) 
		abort(); /* FIXME */
	if (ed_list) {
		gl_list_iterator_t itr;
		const void *p;
		
		itr = gl_list_iterator(ed_list);
		while (gl_list_iterator_next(&itr, &p, NULL)) {
			const struct ed_item *item = p;
			struct id3_frame *frame;
			int i = 0;
			
			while (frame = id3_tag_findframe(tag, item->id, i)) {
				if (ed_item_matches_frame(item, frame)) {
					id3_tag_detachframe(tag, frame);
					id3_frame_delete(frame);
				} else
					i++;
			}
		}
		gl_list_iterator_free(&itr);
	} else
		id3_tag_clearframes(tag);
	safe_id3_file_update_and_close(file);
}


char *
idest_ucs4_cvt(id3_ucs4_t const *ucs4)
{
	if (latin1_option) 
		return (char*)id3_ucs4_latin1duplicate(ucs4);
	else
		return (char*)id3_ucs4_utf8duplicate(ucs4);
}

static void
add_stringlist(gl_list_t list, union id3_field *field, int isgenre,
	       size_t *psize)
{
	unsigned i, nstrings = id3_field_getnstrings(field);
	size_t size = 0;
	
	for (i = 0; i < nstrings; i++) {
		id3_ucs4_t const *ucs4;
		char *str;
		
		ucs4 = id3_field_getstrings(field, i);
		if (!ucs4)
			continue;
		if (isgenre)
			ucs4 = id3_genre_name(ucs4);
		str = idest_ucs4_cvt(ucs4);
		size += strlen(str);
		gl_list_add_last(list, str);
	}
	if (psize)
		*psize = size;
}

char *
field_to_string(union id3_field *field, int isgenre)
{
	id3_ucs4_t const *ucs4;
	char *ret = NULL;
	char buf[128];
	
	switch (id3_field_type(field)) {
	case ID3_FIELD_TYPE_TEXTENCODING:
	case ID3_FIELD_TYPE_INT8:
	case ID3_FIELD_TYPE_INT16:
	case ID3_FIELD_TYPE_INT24:
	case ID3_FIELD_TYPE_INT32:
	case ID3_FIELD_TYPE_INT32PLUS:
		snprintf(buf, sizeof(buf), "%ld", field->number.value);
		ret = xstrdup(buf);
		break;
		
	case ID3_FIELD_TYPE_LATIN1:
	case ID3_FIELD_TYPE_LATIN1FULL:
		/* FIXME */
		ret = xstrdup(id3_field_getlatin1(field));
		break;
		
	case ID3_FIELD_TYPE_LATIN1LIST:
		/* FIXME */
		break;
		
	case ID3_FIELD_TYPE_STRING:
		ucs4 = id3_field_getstring(field);;
		if (ucs4)
			ret = idest_ucs4_cvt(ucs4);
		break;
		
	case ID3_FIELD_TYPE_STRINGFULL:
		ucs4 = id3_field_getfullstring(field);
		ret = idest_ucs4_cvt(ucs4);
		break;

	case ID3_FIELD_TYPE_STRINGLIST: {
		gl_list_t list;
		size_t sz;
		gl_list_iterator_t itr;
		const void *p;
		
		list = new_string_list(true);
		add_stringlist(list, field, isgenre, &sz);
		ret = xmalloc(sz + 1);
		ret[0] = 0;
		itr = gl_list_iterator(list);
		while (gl_list_iterator_next(&itr, &p, NULL))
			strcat(ret, p);
		gl_list_iterator_free(&itr);
		gl_list_free(list);
		break;
	}
		
	case ID3_FIELD_TYPE_LANGUAGE:
	case ID3_FIELD_TYPE_FRAMEID:
	case ID3_FIELD_TYPE_DATE:
		ret = xstrdup(field->immediate.value);
		break;
		
	case ID3_FIELD_TYPE_BINARYDATA:
		/* FIXME */
		break;
	}
	return ret;
}

static void
add_field(gl_list_t list, struct id3_frame *frame, union id3_field *field)
{
	id3_ucs4_t const *ucs4;
	char *str;
	
	switch (id3_field_type(field)) {
	case ID3_FIELD_TYPE_TEXTENCODING:
		break;
	case ID3_FIELD_TYPE_LATIN1:
	case ID3_FIELD_TYPE_LATIN1FULL:
	case ID3_FIELD_TYPE_LATIN1LIST:
		break;
	case ID3_FIELD_TYPE_STRING:
		break;
	case ID3_FIELD_TYPE_STRINGFULL:
		ucs4 = id3_field_getfullstring(field);
		str = idest_ucs4_cvt(ucs4);
		gl_list_add_last(list, str);
		break;
	case ID3_FIELD_TYPE_STRINGLIST:
		add_stringlist(list, field,
			       strcmp(frame->id, ID3_FRAME_GENRE) == 0,
			       NULL);
		break;
	case ID3_FIELD_TYPE_LANGUAGE:
	case ID3_FIELD_TYPE_FRAMEID:
	case ID3_FIELD_TYPE_DATE:
	case ID3_FIELD_TYPE_INT8:
	case ID3_FIELD_TYPE_INT16:
	case ID3_FIELD_TYPE_INT24:
	case ID3_FIELD_TYPE_INT32:
	case ID3_FIELD_TYPE_INT32PLUS:
	case ID3_FIELD_TYPE_BINARYDATA:;
	}
}

void
ed_item_set_comment_fields(struct ed_item *itm, struct id3_frame *frame)
{
	union id3_field *field;
	
	if (strcmp(frame->id, ID3_FRAME_COMMENT) == 0) {
		field = id3_frame_field(frame, 1);
		if (!field)
			return;
		itm->lang = field_to_string(field, 0);
		field = id3_frame_field(frame, 2);
		if (!field)
			return;
		itm->condesc = field_to_string(field, 0);
	}
}

static gl_list_t 
frame_to_list(struct id3_frame *frame)
{
	gl_list_t list;
	unsigned i;
	union id3_field *field;

	list = new_string_list(true);
	for (i = 0; (field = id3_frame_field(frame, i)); i++) 
		add_field(list, frame, field);
	return list;
}

static void
show_tags(struct id3_tag *tag)
{
	struct id3_frame *frame;
	unsigned i;

	for (i = 0; (frame = id3_tag_findframe(tag, NULL, i)); i++) {
		gl_list_t list = frame_to_list(frame);
		if (gl_list_size(list) > 0)
			ed_list_add_item(frame, list);
		else
			gl_list_free(list);
	}
	ed_list_print();
	ed_list_clear();
}

void
query_tags(const char *name)
{
	struct id3_file *file;
	struct id3_tag *tag;
	
	file = id3_file_open(name, ID3_FILE_MODE_READONLY);
	if (!file)
		error(1, errno, "cannot open file %s", name);

	tag = id3_file_tag(file);
	if (tag) {
		if (guile_list(name, tag) == 0)
			show_tags(tag);
	}
	
	id3_file_close(file);
}

static int
prinfo(struct id3_tag *tag, unsigned long location,
       id3_length_t length, void *data)
{
	unsigned int ver = id3_tag_version(tag);
	unsigned int major = ID3_TAG_VERSION_MAJOR(ver);
	printf("version: %u.%u\n",
	       major > 1 ? 20 + major : major,
	       ID3_TAG_VERSION_MINOR(ver));
	printf("offset: %lu\n", location);
	printf("length: %lu\n", length);
	return 0;
}

void
info_id3(const char *name)
{
	struct id3_file *file;
	unsigned long ntags;
	
	file = id3_file_open(name, ID3_FILE_MODE_READONLY);
	if (!file)
		error(1, errno, "cannot open file %s", name);
	ntags = id3_file_struct_ntags(file);
	printf("file: %s\n", name);
	printf("ntags: %lu\n", ntags);
	id3_file_struct_iterate(file, prinfo, NULL);
	id3_file_close(file);
}


